use bitvec::prelude::*;
use goertzel_nostd::Parameters;
use std::collections::VecDeque;
use std::iter::Iterator;

pub struct DownSampleIter<I, T: Iterator<Item = I>> {
    source: T,
    scale: f32,
    sample_i: f32,
}

impl<I, T: Iterator<Item = I>> DownSampleIter<I, T> {
    pub fn new(source: T, source_sample_rate: f32, dest_sample_rate: f32) -> Self {
        assert!(source_sample_rate > dest_sample_rate, "Downsampling only!");

        Self {
            source,
            scale: source_sample_rate / dest_sample_rate,
            sample_i: 0.0,
        }
    }
}

impl<I, T: Iterator<Item = I>> Iterator for DownSampleIter<I, T> {
    type Item = I;

    fn next(&mut self) -> Option<I> {
        for s in self.source.by_ref() {
            self.sample_i += 1.0;

            if self.sample_i > self.scale {
                self.sample_i -= self.scale;
                return Some(s);
            }
        }

        None
    }
}

pub struct DecoderIter<T: Iterator<Item = f32>> {
    source: T,
    sample_rate: f32,
    baud: u32,
    pll: f32,
    bit_samples: VecDeque<f32>,
    prev_nrz: f32,

    upper: Parameters,
    lower: Parameters,
}

impl<T: Iterator<Item = f32>> DecoderIter<T> {
    pub fn new(source: T, sample_rate: f32, baud: u32) -> Self {
        let bit_samples_len = ((sample_rate / baud as f32) + 1.0) as usize;

        let mut bit_samples = VecDeque::new();
        for _ in 0..bit_samples_len {
            bit_samples.push_back(0.0);
        }

        Self {
            source,
            sample_rate,
            baud,
            pll: 0.0,
            bit_samples,
            prev_nrz: 0.0,
            lower: Parameters::new(1200.0, sample_rate as u32, bit_samples_len),
            upper: Parameters::new(2200.0, sample_rate as u32, bit_samples_len),
        }
    }

    // pll
    // based off of direwolf and https://inst.eecs.berkeley.edu/~ee123/sp17/lab/lab5/Lab5_Part_2-Audio_Frequency_Shift_Keying(AFSK).html
    fn sample(&mut self) -> Option<f32> {
        let value = self.decode_window();

        const A: f32 = 0.75;
        let zero_crossing =
            (self.prev_nrz > 0.0 && value < 0.0) || (self.prev_nrz < 0.0 && value > 0.0);
        if zero_crossing {
            self.pll *= A;
        }

        self.pll += self.baud as f32 / self.sample_rate as f32;
        let overflow = self.pll > 0.5;

        // println!("{}\t{}\t{}", value, self.pll, zero_crossing);

        self.prev_nrz = value;
        if overflow {
            self.pll -= 1.0;
            Some(value)
        } else {
            None
        }
    }

    fn decode_window(&mut self) -> f32 {
        self.bit_samples.make_contiguous();
        self.lower.mag_squared(self.bit_samples.as_slices().0)
            - self.upper.mag_squared(self.bit_samples.as_slices().0)
    }
}

impl<T: Iterator<Item = f32>> Iterator for DecoderIter<T> {
    type Item = bool;

    fn next(&mut self) -> Option<bool> {
        while let Some(value) = self.source.next() {
            // slide our window by 1 sample
            self.bit_samples.pop_front();
            self.bit_samples.push_back(value);

            if let Some(res) = self.sample() {
                return Some(res < 0.0);
            }
        }

        None
    }
}

pub struct NrziIter<T: Iterator<Item = bool>> {
    source: T,
    previous: bool,
}

impl<T: Iterator<Item = bool>> NrziIter<T> {
    pub fn new(source: T) -> Self {
        Self {
            source,
            previous: true,
        }
    }
}

impl<T: Iterator<Item = bool>> Iterator for NrziIter<T> {
    type Item = bool;

    fn next(&mut self) -> Option<bool> {
        let b = self.source.next()?;
        if b != self.previous {
            self.previous = b;
            Some(false)
        } else {
            Some(true)
        }
    }
}

// Assumes that a frame will always have a multiple of 8 bits
pub struct FrameIter<T: Iterator<Item = bool>> {
    source: T,
    buffer: u8,
    max_frame_size: usize,
}

impl<T: Iterator<Item = bool>> FrameIter<T> {
    pub fn new(source: T, max_frame_size: usize) -> Self {
        // # bytes -> # bits
        let max_frame_size = max_frame_size * 8;
        Self {
            source,
            buffer: 0,
            max_frame_size,
        }
    }
}

impl<T: Iterator<Item = bool>> Iterator for FrameIter<T> {
    type Item = BitVec<u8, Msb0>;

    fn next(&mut self) -> Option<Self::Item> {
        let mut in_frame = false;
        let mut frame = BitVec::new();

        for x in self.source.by_ref() {
            let b = if x { 1 } else { 0 };

            if self.buffer == 0x7e && !in_frame {
                // start of frame
                in_frame = true;
            }

            if in_frame {
                frame.push(x);

                // If we get a random string of bits that happens to be 0x7e,
                // we need to give up after a certain amount of time not receiving the
                // end flag.
                if frame.len() > self.max_frame_size {
                    in_frame = false;
                    frame.clear();
                }
            }

            self.buffer <<= 1;
            self.buffer |= b;

            if self.buffer == 0x7e && in_frame {
                for _ in 0..8 {
                    frame.pop();
                }

                if frame.is_empty() {
                    in_frame = false;
                } else {
                    return Some(frame);
                }
            }
        }

        None
    }
}
